"""Tests the utopya models <...> subcommands of the CLI"""

import os
import traceback

import pytest

import utopya

from .._fixtures import *
from . import invoke_cli

# -----------------------------------------------------------------------------


def test_list(with_test_models):
    """Tests utopya models ls"""
    # Lists models as expected
    res = invoke_cli(("models", "ls"))
    assert "model registry" in res.output
    assert DUMMY_MODEL in res.output

    # Also shows the number of bundles available in "long" mode
    res = invoke_cli(("models", "ls", "--long"))
    assert "bundle" in res.output
    assert "Default bundles are marked" in res.output
    assert "(*)" in res.output

    assert res.output == invoke_cli(("models", "ls", "-l")).output


def test_register_single(registry):
    """Tests utopya models register single"""
    reg_args = (
        "models",
        "register",
        "single",
        DUMMY_MODEL,
        "-e",
        DUMMY_EXECUTABLE,
    )
    reg_args += ("--label", TEST_LABEL)  # already exists
    reg_args += ("--source-dir", "./")

    res = invoke_cli(reg_args)
    print(res.output)
    assert res.exit_code != 0
    assert "Registration failed!" in res.output
    assert "Bundle validation failed" in res.output


def test_register_from_list(registry):
    """Tests utopya models register single

    NOTE The fixture already performs some of the tests
    """
    reg_args = (
        "models",
        "register",
        "from-list",
        DUMMY_MODEL,
        "--executables",
        DUMMY_EXECUTABLE,
    )
    res = invoke_cli(reg_args + ("--label", "some_new_label"))
    print(res.output)
    if res.exception:
        traceback.print_tb(res.exception.__traceback__)
    assert res.exit_code == 0
    assert "Model registration succeeded" in res.output

    # Can also use format strings
    reg_args = (
        "models",
        "register",
        "from-list",
        DUMMY_MODEL,
        "--executable-fstr",
        "{model_name:}/{model_name:}.py",
        "--source-dir-fstr",
        "{model_name:}/",
        "--base-executable-dir",
        os.path.join(DEMO_DIR, "models"),
        "--base-source-dir",
        os.path.join(DEMO_DIR, "models"),
    )
    res = invoke_cli(reg_args + ("--label", "yet_another_label"))
    print(res.output)
    assert res.exit_code == 0
    assert "Model registration succeeded" in res.output

    # Mutually exclusive
    res = invoke_cli(
        (
            "models",
            "register",
            "from-list",
            DUMMY_MODEL,
            "--executable-fstr",
            "{model_name:}/{model_name:}.py",
            "--source-dir-fstr",
            "{model_name:}/",
            "--source-dirs",
            "/foo/bar",
        )
    )
    print(res.output)
    assert res.exit_code != 0
    assert "mutually exclusive" in res.output

    # Superfluous arguments
    res = invoke_cli(reg_args + ("--executables", "'foo;bar'"))
    print(res.output)
    assert res.exit_code != 0
    assert "mutually exclusive" in res.output

    # Missing arguments
    reg_args = (
        "models",
        "register",
        "from-list",
        DUMMY_MODEL,
    )
    res = invoke_cli(reg_args)
    print(res.output)
    assert res.exit_code != 0
    assert "Missing argument --executables or" in res.output

    # Length mismatch
    reg_args = (
        "models",
        "register",
        "from-list",
        f"'{DUMMY_MODEL};{DUMMY_MODEL}'",
        "--executables",
        DUMMY_EXECUTABLE,
    )
    res = invoke_cli(reg_args)
    print(res.output)
    assert res.exit_code != 0
    assert "Mismatch of sequence lengths" in res.output


def test_register_from_manifest(registry):
    """Tests utopya models register from-manifest"""

    DUMMY_INFO = os.path.join(
        DEMO_DIR, "models", DUMMY_MODEL, f"{DUMMY_MODEL}_info.yml"
    )
    reg_args = ("models", "register", "from-manifest", DUMMY_INFO)
    reg_args += ("--model-name", DUMMY_MODEL)

    res = invoke_cli(reg_args)
    print(res.output)
    assert res.exit_code == 0
    assert "from_manifest_file" in res.output

    # With custom label
    extd_reg_args = reg_args + ("--label", "custom_label")
    res = invoke_cli(extd_reg_args)
    print(res.output)
    assert res.exit_code == 0
    assert "from_manifest_file" not in res.output
    assert "custom_label" in res.output
    assert "custom_label" in registry[DUMMY_MODEL]

    # With custom model name
    extd_reg_args += ("--model-name", "MyCustomModelName")
    res = invoke_cli(extd_reg_args)
    print(res.output)
    assert len(registry["MyCustomModelName"]) == 1
    assert "custom_label" in registry["MyCustomModelName"]
    registry.remove_entry("MyCustomModelName")

    # Custom model name fails with more than one manifest file
    res = invoke_cli(extd_reg_args + (DUMMY_INFO,))
    print(res.output)
    assert res.exit_code != 0
    assert "can only be specified if only a single manifest file" in res.output


def test_remove(registry):
    """Tests utopya models rm

    This is based on the registry fixture, which adds a bunch of entries to
    the model registry which can then be removed here.
    """
    # Add some additional entries (to have something to remove)
    labels = (
        f"{TEST_LABEL}_foo",
        f"{TEST_LABEL}_bar",
        f"{TEST_LABEL}_spam",
    )
    for label in labels:
        reg_args = (
            "models",
            "register",
            "single",
            DUMMY_MODEL,
            "-e",
            DUMMY_EXECUTABLE,
        )
        reg_args += ("--label", label)
        reg_args += ("--source-dir", "./")

        res = invoke_cli(reg_args)
        print(res.output)
        assert res.exit_code == 0

    # Remove a single label
    res = invoke_cli(("models", "rm", DUMMY_MODEL, "--label", labels[0]))
    assert res.exit_code == 0
    assert TEST_LABEL in registry[DUMMY_MODEL]
    assert labels[0] not in registry[DUMMY_MODEL]
    assert labels[1] in registry[DUMMY_MODEL]
    assert labels[2] in registry[DUMMY_MODEL]

    # ... with prompt for bundle name
    res = invoke_cli(("models", "rm", DUMMY_MODEL), input=f"{labels[1]}\n")
    assert res.exit_code == 0
    assert TEST_LABEL in registry[DUMMY_MODEL]
    assert labels[0] not in registry[DUMMY_MODEL]
    assert labels[1] not in registry[DUMMY_MODEL]
    assert labels[2] in registry[DUMMY_MODEL]

    # Now remove it completely
    res = invoke_cli(("models", "rm", DUMMY_MODEL, "--all"), input="N\n")
    print(res.output)
    assert res.exit_code == 0
    assert "Not removing anything" in res.output

    res = invoke_cli(("models", "rm", DUMMY_MODEL, "--all"), input="y\n")
    print(res.output)
    assert res.exit_code == 0
    assert DUMMY_MODEL not in registry


def test_set_default(registry):
    """Tests utopya models set-default"""
    assert registry[DUMMY_MODEL].default_label == TEST_LABEL

    # Unset default and re-set via CLI
    registry[DUMMY_MODEL].default_label = None
    res = invoke_cli(("models", "set-default", DUMMY_MODEL, TEST_LABEL))
    print(res.output)
    assert res.exit_code == 0
    assert TEST_LABEL in res.output
    assert registry[DUMMY_MODEL].default_label == TEST_LABEL

    # Fails for invalid label name, keeping the old default
    assert "invalid_label" not in registry[DUMMY_MODEL]
    res = invoke_cli(("models", "set-default", DUMMY_MODEL, "invalid_label"))
    print(res.output)
    assert res.exit_code != 0
    assert "invalid_label" in res.output
    assert registry[DUMMY_MODEL].default_label == TEST_LABEL


def test_edit(monkeypatch, with_test_models):
    """Tests utopya models edit"""
    res = invoke_cli(("models", "edit", DUMMY_MODEL), input="y\n")
    print(res.output)
    assert res.exit_code == 1
    assert "Editing model registry file failed!" in res.output

    monkeypatch.setenv("EDITOR", "echo")  # this will always work
    res = invoke_cli(("models", "edit", DUMMY_MODEL), input="y\n")
    print(res.output)
    assert res.exit_code == 0
    assert "Successfully edited registry file" in res.output

    # Not continuing
    res = invoke_cli(("models", "edit", DUMMY_MODEL), input="N\n")
    print(res.output)
    assert res.exit_code == 0
    assert "Not opening" in res.output


@pytest.mark.skip("NotImplemented")
def test_copy():
    """Tests utopya models copy"""
    new_model_args = ("--new-name", "foo", "--target-project", "bar")
    res = invoke_cli(("models", "copy", DUMMY_MODEL) + new_model_args)
    assert res.exit_code == 0



"""Implements the `utopya projects` subcommand"""

import os
import sys

import click

from ._utils import Echo

projects = click.Group(
    name="projects",
    help="Show available projects and register new ones.",
)


# .. utopya projects ls .......................................................


@projects.command(
    name="ls",
    help="Lists all registered projects.",
)
@click.option(
    "-l",
    "--long",
    "long_mode",
    is_flag=True,
    help="Show more detailed information.",
)
def list_projects(long_mode: bool):
    """Lists available projects"""
    Echo.info("Loading utopya project list ...")

    from utopya import PROJECTS
    from utopya.tools import pformat

    Echo.info("\n--- Utopya Projects ---")
    for project_name, project in PROJECTS.items():
        Echo.info(f"- {project_name}")
        if not long_mode:
            continue

        for k, v in project.data.dict().items():
            Echo.remark(f"  {k:15s}: {v}")  # TODO make dict-compatible
        Echo.info("")


# .. utopya projects edit .....................................................


@projects.command(help="Edit a project's registry file directly.")
@click.argument("project_name")
def edit(project_name: str):
    """Edits a project registry file"""
    Echo.info(f"Editing registry file for project '{project_name}' ...")
    Echo.warning("Take care not to corrupt the file!")

    if not click.confirm("Open file for editing?"):
        Echo.info("Not opening.")
        sys.exit(0)

    try:
        from utopya import PROJECTS

        click.edit(
            filename=PROJECTS[project_name].registry_file_path,
            extension=".yml",
        )

    except Exception as exc:
        Echo.error("Editing project registry file failed!", error=exc)
        sys.exit(1)

    Echo.success(f"Successfully edited project registry file.")


# .. utopya projects rm .......................................................


@projects.command(
    name="rm",
    help="Remove a project.",
)
@click.argument("project_name")
@click.option(
    "-y",
    "skip_confirmation",
    is_flag=True,
    help="If given, will skip the confirmation prompt.",
)
def remove(
    *,
    project_name: str,
    skip_confirmation: bool,
):
    """Removes an entry from the project registry file"""
    from utopya import PROJECTS

    Echo.info(f"Removing project registry entry '{project_name}' ...")

    if not skip_confirmation and not click.confirm("Are you sure?"):
        Echo.info("Not removing.")
        sys.exit(0)

    if project_name not in PROJECTS:
        Echo.error(f"There is no project named '{project_name}' registered!")
        Echo.info(f"Registered projects:  {', '.join(projects)}")
        sys.exit(1)

    PROJECTS.remove_entry(project_name)
    Echo.success(f"Successfully removed project entry '{project_name}'.")


# .. utopya projects register .................................................


@projects.command(
    name="register",
    help=(
        "Register a project or validate an existing one.\n"
        "\n"
        "Required arguments are the base directory of the project."
    ),
)
@click.argument(
    "base_dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
)
@click.option(
    "--info-file",
    default=None,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help=(
        "A file that contains all additional project information like "
        "the project name, relevant paths, or other metadata. "
        "If not given, will search for some candidate files relative to "
        "BASE_DIR, e.g. `project_info.yml`."
    ),
)
@click.option(
    "--custom-name",
    "custom_project_name",
    type=str,
    help=(
        "A custom project name that may differ from the one given in the "
        "project info file."
    ),
)
@click.option(
    "--require-matching-names",
    is_flag=True,
    help=(
        "If set, requires that an optionally given --custom-name and the name "
        "set in the project info file match exactly."
    ),
)
@click.option(
    "--exists-action",
    default="validate",
    type=click.Choice(("raise", "validate", "overwrite", "update")),
    help="What to do if a project of the same name is already registered.",
)
def register(**kwargs):
    """Registers a project or validates an existing one"""
    from utopya import PROJECTS

    try:
        PROJECTS.register(**kwargs)

    except Exception as exc:
        Echo.error("Project registration failed!", error=exc)
        raise
        sys.exit(1)



"""Implements a YAML-based registry infrastructure"""

import copy
import logging
import os

import dantro.utils

from ..exceptions import (
    EntryExistsError,
    EntryValidationError,
    MissingEntryError,
)
from ..tools import make_columns, pformat, recursive_update
from .entry import RegistryEntry

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class KeyOrderedDict(dantro.utils.KeyOrderedDict):
    """A key-ordered dict that expects string keys and sorts by the lower-case
    representation of keys.
    """

    DEFAULT_KEY_COMPARATOR = lambda _, k: k.lower()


# -----------------------------------------------------------------------------


class YAMLRegistry:
    """A registry framework that persistently stores the registry entries as
    YAML files within a common directory.

    Individual registry entries can be retrieved via a dict-like interface.
    """

    def __init__(self, EntryCls: type, *, registry_dir: str):
        """Set up a registry directory for a certain class of registry entries.

        Args:
            EntryCls (type): Type of the individual entries
            registry_dir (str): Path to the directory in which the individual
                registry entry files are to be stored.
        """
        if not issubclass(EntryCls, RegistryEntry):
            raise TypeError(
                f"EntryCls needs to be a subclass of {RegistryEntry}, "
                f"but was {EntryCls}!"
            )

        self._registry_dir = registry_dir
        self._EntryCls = EntryCls

        self._registry = KeyOrderedDict()
        self.reload()

    @property
    def registry_dir(self) -> str:
        """The associated registry directory"""
        return self._registry_dir

    def __str__(self) -> str:
        return "<{}, entry type: {}, @ {} >".format(
            type(self).__name__, self._EntryCls.__name__, self.registry_dir
        )

    def reload(self):
        """Load all available entries from the registry directory.

        If called multiple times, will only load entries that are not already
        loaded.
        """
        log.debug("Disassociating existing entries ...")
        for entry in self.values():
            entry._registry = None
        self._registry = KeyOrderedDict()

        log.debug("Re-loading entries from registry directory ...")
        new_entries = []
        for fname in os.listdir(self.registry_dir):
            name, ext = os.path.splitext(fname)

            if name in self or ext != self._EntryCls.FILE_EXTENSION:
                continue

            entry = self._EntryCls(name=name, registry=self)
            self._registry[entry.name] = entry
            new_entries.append(name)

        log.debug(
            "Loaded %s entr%s: %s",
            len(new_entries),
            "ies" if len(new_entries) != 1 else "y",
            ", ".join(new_entries),
        )

    # Dict interface ..........................................................

    def keys(self):
        return self._registry.keys()

    def values(self):
        return self._registry.values()

    def items(self):
        return self._registry.items()

    def __iter__(self):
        return self._registry.__iter__()

    def __contains__(self, name: str) -> bool:
        """Whether an entry of the given name exists in the registry"""
        return name in self._registry

    def __len__(self) -> int:
        return len(self._registry)

    # Working on entries ......................................................

    def __getitem__(self, name: str):
        """Retrieve a deep copy of a model registration entry for the given
        model name.
        """
        try:
            return self._registry[name]

        except KeyError as err:
            raise MissingEntryError(
                f"{self._EntryCls.__name__} '{name}' not found in {self}! "
                f"Available entries:\n{make_columns(self.keys())}"
            ) from err

    def __delitem__(self, name: str):
        """Removes a registry entry"""
        self.remove_entry(name)

    # Adding and removing .....................................................

    def add_entry(
        self, name: str, *, exists_action: str = "raise", **data
    ) -> RegistryEntry:
        """Creates a new entry and stores it in the registry. If an entry of
        the same name already exists, allows according to the ``exists_action``
        Adds a new entry of a certain name; raises if it already exists.

        TODO Write

        Args:
            name (str): Description
            exists_action (str, optional): Description
            **data: Description

        Returns:
            RegistryEntry: Description

        Raises:
            EntryExistsError: Description
            ValidationError:
            ValueError:
        """
        # Construct the entry itself, but *without* associating it with the
        # registry -- this allows to evaluate the exists action:
        new_entry = self._EntryCls(name=name, registry=None, **data)

        if name in self:
            log.caution("An entry named '%s' already exists!", name)

            if exists_action == "raise":
                raise EntryExistsError(
                    f"An entry '{name}' in {self} already exists! "
                    "Either remove it or choose a different `exists_action`."
                )

            elif exists_action == "validate":
                log.remark("Validating new entry against existing entry ...")
                if self[name] != new_entry:
                    # Generate a diff such that its clearer where they differ
                    import difflib
                    import json

                    natify = lambda d: json.loads(d.json())

                    diff = "\n".join(
                        difflib.Differ().compare(
                            pformat(natify(self[name])).split("\n"),
                            pformat(natify(new_entry)).split("\n"),
                        )
                    )

                    raise EntryValidationError(
                        f"Validation of project '{name}' failed!\n"
                        "The to-be-added project information did not compare "
                        "equal to the already existing one for that project.\n"
                        "Either change the `exists_action` argument to "
                        "'overwrite' or 'update' or make sure the information "
                        f"is equal.\nTheir YAML diff is as follows:\n\n{diff}"
                    )

                # else: no need to change anything below
                log.remark("Validation of entry '%s' succeeded.", name)
                return

            elif exists_action == "update":
                log.remark("Updating existing entry with new entry ...")
                data = recursive_update(self[name].dict(), copy.deepcopy(data))
                new_entry = self._EntryCls(name=name, registry=None, **data)

            elif exists_action == "overwrite":
                log.remark("Overwriting already existing entry ...")
                pass

            elif exists_action == "skip":
                log.remark("Not adding the new entry.")
                return

            else:
                raise ValueError(
                    f"Invalid `exists_action` '{exists_action}'!\n"
                    "Possible values: raise, validate, update, overwrite, skip"
                )

        # Now, make the registry association, store it here and write the file
        new_entry._set_registry(self)
        new_entry.write()
        self._registry[new_entry.name] = new_entry

        log.success("Added entry '%s'.", new_entry.name)
        return new_entry

    def remove_entry(self, name: str):
        """Removes a registry entry and deletes the associated registry file.

        Args:
            name (str): The name of the entry that is to be removed

        Raises:
            ValueError: On invalid (non-existing) model
        """
        try:
            entry = self._registry.pop(name)

        except KeyError as err:
            raise MissingEntryError(
                f"Could not remove registry entry '{name}', because "
                "no such entry is present.\nAvailable entries:\n"
                f"{make_columns(self.keys())}"
            ) from err
        else:
            log.debug("Removed entry '%s' from %s.", name, self)

        entry.remove_registry_file()
        log.debug(
            "Removed associated registry file:  %s", entry.registry_file_path
        )
        entry._registry = None
        # Entry goes out of scope now and is then be garbage-collected if it
        # does not exist anywhere else... Only if some action is taken on that
        # entry does it lead to file being created again.



"""Implements the utopya eval CLI subtree"""

import copy
import logging
import readline
import sys
import traceback
from typing import List, Tuple, Union

import click

from ._shared import (
    INTERACTIVE_MODE_PROHIBITED_ARGS,
    OPTIONS,
    add_options,
    default_none,
)
from ._utils import ANSIesc, Echo, parse_run_and_plots_cfg, parse_update_dicts

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


@click.command(
    name="eval",
    help=(
        "Evaluate a previously finished simulation run.\n"
        "\n"
        "Loads a simulation of the given MODEL_NAME model and evaluates it "
        "either using the configured defaults or with custom plots configs. "
        "If no RUN_DIR is given, will use the latest output; to evaluate a "
        "specific simulation, the directory name can be used."
    ),
)
@click.argument("model_name")
@click.argument(
    "run_dir",
    required=False,
)
@click.option(
    "--run-cfg",
    default=None,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help=(
        "An optional run config that can be used if changes to the meta "
        "config are desired. This affects evaluation indirectly as it "
        "determines some of the behavior for data loading and setup of the "
        "plotting infrastructure."
    ),
)
@add_options(OPTIONS["model_selection"])
@click.option(
    "--set-params",
    "-p",
    multiple=True,
    callback=default_none,
    help=(
        "Sets entries in the meta configuration using key-value pairs. "
        "Example: -p some.param=42 sets the 'param' entry in 'some' to 42. "
        "Specify DELETE as value to remove an entry. "
        "Repeat the -p option to set multiple values."
    ),
)
#
# -- Evaluation
#
@add_options(OPTIONS["load"])
@add_options(OPTIONS["eval"])
@add_options(OPTIONS["debug_flag"])  # --debug
@add_options(OPTIONS["cluster_mode"])  # --cluster-mode
#
#
#
@click.pass_context
def evaluate(ctx, **params):
    """Invokes a model simulation run and subsequent evaluation"""
    import utopya
    from utopya.exceptions import ValidationError
    from utopya.tools import pformat

    _log = utopya._getLogger("utopya")  # TODO How best to do this?!

    # Preparations . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    _log.info("Parsing additional command line arguments ...")
    update_dict, update_plots_cfg = parse_update_dicts(_mode="eval", **params)

    if update_dict:
        _log.note("Updates to meta configuration:\n\n%s", pformat(update_dict))

    model = utopya.Model(
        name=params["model_name"],
        bundle_label=params["label"],
    )

    run_cfg, plots_cfg = parse_run_and_plots_cfg(
        model=model,
        run_cfg=params["run_cfg"],
        plots_cfg=params["plots_cfg"],
        cfg_set=params["cfg_set"],
        _log=_log,
    )
    params["plots_cfg"] = plots_cfg
    params["update_plots_cfg"] = update_plots_cfg

    # Create frozen Multiverse; supplies similar interface as Multiverse
    mv = model.create_frozen_mv(
        run_dir=params["run_dir"], run_cfg_path=run_cfg, **update_dict
    )

    # Loading and evaluating . . . . . . . . . . . . . . . . . . . . . . . . .
    _load_and_eval(
        _log=_log,
        ctx=ctx,
        mv=mv,
        **params,
    )


# -----------------------------------------------------------------------------


def _load_and_eval(
    *,
    ctx,
    mv: Union["Multiverse", "FrozenMultiverse"],
    _log=log,
    **params,
):
    """Wrapper that takes care of loading and evaluating"""
    _log.progress(
        "Beginning evaluation: loading data and starting PlotManager ...\n"
    )

    # Loading data into the data tree and (optionally) showing it .............
    if not params["use_data_tree_cache"]:
        mv.dm.load_from_cfg()

    else:
        if not mv.dm.tree_cache_exists:
            mv.dm.load_from_cfg()
            mv.dm.dump()

        else:
            _log.hilight("Restoring tree from cache file ...")
            mv.dm.restore()

    if params["show_data_tree"] == "full":
        _log.info(mv.dm.tree)

    elif params["show_data_tree"] == "condensed":
        _log.info(mv.dm.tree_condensed)

    # Plotting ................................................................
    if not params["interactive"]:
        mv.pm.plot_from_cfg(
            plots_cfg=params["plots_cfg"],
            plot_only=params["plot_only"],
            **params["update_plots_cfg"],
        )

        if params["reveal_output"] and mv.pm.common_out_dir:
            _log.progress("Revealing output ...")
            _log.remark("Output directory:\n  %s", mv.pm.common_out_dir)
            click.launch(mv.pm.common_out_dir)

        _log.success("All done.\n")
        return

    # Interactive Plotting ....................................................
    # Carry over all args relevant for evaluation.
    args = sys.argv[2:]

    # For convenience, drop a potentially given run directory path argument,
    # which would have to be removed from the argument list manually otherwise.
    args = [arg for arg in args if arg not in ()]

    # ... and drop some other flags and combinations of shortcuts of flags
    # that will no longer be necessary.
    # NOTE This is a rather rudimentary approach and will probably not remove
    #      *all* possibly offending arguments. It's purely for convenience at
    #      this point. If there are remaining args that are not parseable, the
    #      parser in interactive plotting mode will complain and the input can
    #      be adjusted at that point.
    # TODO Do the inverse (white-listing) approach instead!
    args = [
        arg
        for arg in args
        if arg
        not in (
            #
            # Options
            "--cluster-mode",
            "--interactive",
            "-i",
            "--use-data-tree-cache",
            "--tc",
            "--show-data-tree",
            "--load-parallel",
            "-P",
            #
            # Combinations
            "-iP",
            "-Pi",
            #
            # Values
            ctx.params.get("run_cfg"),
            ctx.params.get("run_dir"),
            ctx.params["show_data_tree"],
        )
    ]

    # While in interactive mode, the following loop is carried out repeatedly.
    # A counter variable is used to label successive output directories.
    iap_session = 0

    while params["interactive"]:
        print("")
        _log.hilight("--- Interactive plotting session %d ---", iap_session)

        # Unless this is the very first session, need to prompt for new args.
        if iap_session > 0:
            # Provide some information on where the data is from; which is
            # useful if this is a very long session
            _log.remark(
                "Currently selected data directory:\n  %s", mv.dm.dirs["data"]
            )
            _log.note("Use Ctrl+C to exit. Use ↑/↓ for history.")

            # Prompt for new arguments . . . . . . . . . . . . . . . . . . . .
            try:
                args, params, ctx = _prompt_new_params(
                    old_args=args,
                    old_params=params,
                    old_ctx=ctx,
                    _log=_log,
                )

            except KeyboardInterrupt:
                # Ask for confirmation before quitting the plotting session.
                # This is to prohibit accidentally exiting the session, which
                # is especially relevant if loading data took a long time.
                print("\n")
                readline.set_startup_hook()
                _prompt = (
                    "{ansi.ORANGE}{ansi.BOLD}"
                    "Really exit interactive plotting session?"
                    "{ansi.RESET}"
                ).format(ansi=ANSIesc)

                try:
                    if click.confirm(_prompt):
                        break

                except (KeyboardInterrupt, click.exceptions.Abort):
                    pass

                print("\n")
                _log.success("Remaining in interactive plotting mode ...")
                continue

            except EOFError:
                # EOFError is invoked by builtins.input when the input is read
                # from a stream that is not the standard input stream. In such
                # a case, can't have a confirmation, so might as well exit.
                break

            except (SystemExit, click.exceptions.Exit):
                # This came from SystemExit or similar being raised during
                # prompting for new arguments, e.g. because of the --help flag.
                # ... just prompt again; error message was already shown
                continue

            except click.exceptions.UsageError as exc:
                _handle_interactive_plotting_exception(
                    exc,
                    _log=_log,
                    description="command line argument parsing",
                )
                continue

        # Get updated configurations . . . . . . . . . . . . . . . . . . . . .
        try:
            _, plots_cfg = parse_run_and_plots_cfg(
                model=mv.model,
                run_cfg=params["run_cfg"],
                plots_cfg=params["plots_cfg"],
                cfg_set=params["cfg_set"],
                _log=_log,
                _interactive_mode=True,
            )
            _, update_plots_cfg = parse_update_dicts(_mode="eval", **params)

            params["plots_cfg"] = plots_cfg
            params["update_plots_cfg"] = update_plots_cfg

        except Exception as exc:
            _handle_interactive_plotting_exception(
                exc,
                _log=_log,
                debug=params["debug"],
                description="parsing of new PlotManager arguments",
            )
            continue

        if update_plots_cfg:
            _log.note(
                "Updates to plot configuration:\n\n%s",
                pformat(update_plots_cfg),
            )

        # Create a new PlotManager . . . . . . . . . . . . . . . . . . . . . .
        # Uses a custom output directory (inside the regular eval directory)
        # and include the session number into it.
        # Also increments the session counter.
        try:
            mv.renew_plot_manager(
                out_dir=f"session{iap_session:03d}/",
                raise_exc=(params["debug"] >= 1),
            )

        except Exception as exc:
            _handle_interactive_plotting_exception(
                exc,
                _log=_log,
                debug=params["debug"],
                description="PlotManager renewal",
                remark=(
                    "Inspect the traceback for details and check that all "
                    "involved plot configuration files are using valid YAML."
                ),
            )
            continue

        finally:
            iap_session += 1
            print("")

        # Plotting . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        try:
            mv.pm.plot_from_cfg(
                plots_cfg=plots_cfg,
                plot_only=params["plot_only"],
                **update_plots_cfg,
            )

        except KeyboardInterrupt:
            print("")
            _log.caution("Interrupted current plotting session.")
            _log.progress("Remaining in interactive plotting mode ...")
            continue

        except Exception as exc:
            _handle_interactive_plotting_exception(
                exc,
                _log=_log,
                debug=params["debug"],
                description="interactive plotting",
            )
            continue

        # Reveal output . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        if params["reveal_output"] and mv.pm.common_out_dir:
            try:
                click.launch(mv.pm.common_out_dir)

            except Exception as exc:
                # This may fail for certain paths; it should not lead to the
                # plotting session being interrupted.
                _handle_interactive_plotting_exception(
                    exc,
                    _log=_log,
                    debug=params["debug"],
                    description="opening of output directory",
                )
                continue

        # End of interactive plotting loop . . . . . . . . . . . . . . . . . .

    print("\n")
    _log.success("Left interactive plotting session.\n")


def _handle_interactive_plotting_exception(
    exc: Exception,
    *,
    description: str,
    debug: int = None,
    remark: str = None,
    _log=log,
):
    """Helper function to handle exceptions during interactive plotting

    Args:
        exc (Exception): The exception to handle
        description (str): A context-like description that will be used in the
            log message
        debug (int, optional): The debug level. If this evaluates to True,
            a traceback will be shown. If given, regardless of value, there
            will be a note-level message that hints at the possibility of
            showing a traceback; if not given (i.e.: None, default), no such
            message will be shown.
        remark (str, optional): A remark-level message added after the error
            message was provided. Use this to hint at possible ways to address
            the error.
        _log (TYPE, optional): A logger-like instance to use for emitting all
            information.
    """
    if debug:
        _log.error("An error occured during %s!\n", description)
        traceback.print_exc()
        _log.note("Remove --debug flag to hide traceback.")

    else:
        _log.error(
            "An error occured during %s!\n\n%s%s%s: %s\n",
            description,
            ANSIesc.BOLD,
            exc.__class__.__name__,
            ANSIesc.RESET + ANSIesc.RED,
            str(exc),
        )
        _log.note("Add --help to show usage info.")
        if debug is not None:
            _log.note("Add --debug to show traceback.")

    if remark:
        _log.remark(remark)
    _log.progress("Remaining in interactive plotting mode ...")


def _prompt_new_params(
    *,
    old_args: List[str],
    old_params: dict,
    old_ctx: click.Context,
    _log=log,
) -> Tuple[list, dict, click.Context]:
    """Given some old arguments, prompts for new ones and returns a new
    list of argument values and the parsed argparse namespace result.

    Args:
        old_args (List[str]): The old argument value list
        old_params (dict): The old set of parsed arguments
        old_ctx (click.Context): The CLI context
        _log: A logger-like object

    Returns:
        Tuple[list, dict]: New argument value list and parsed arguments.

    Raises:
        ValueError: Upon error in parsing the new arguments.
    """
    # Create a new argument list for querying the user. For that, remove
    # those entries from the argss that are meant to be in the query.
    prefix_args = ("--interactive", old_params["model_name"])
    to_query = [arg for arg in old_args if arg not in prefix_args]
    to_query_str = " ".join(to_query) + (" " if to_query else "")

    # Now, setup the startup hook with a callable that inserts those
    # arguments that shall be editable by the user. Configure readline to
    # allow tab completion for file paths after certain delimiters.
    readline.set_startup_hook(lambda: readline.insert_text(to_query_str))
    readline.parse_and_bind("tab: complete")
    readline.set_completer_delims(" \t\n=")

    # Generate the prompt and store the result, stripping whitespace
    prompt_str = (
        "\n{ansi.CYAN}${ansi.MAGENTA} "
        "utopya eval -i {}"
        "{ansi.RESET} ".format(old_params["model_name"], ansi=ANSIesc)
    )
    input_res = input(prompt_str).strip()
    print("")

    # Reset the startup hook to do nothing
    readline.set_startup_hook()

    # Prepare the new list of arguments
    add_args = input_res.split(" ") if input_res else []
    new_args = list(prefix_args) + add_args

    # Now let the click.Context create a new context, parsing the arguments
    # NOTE This may raise SystemExit or click.exceptions.Exit upon the --help
    #      argument or click.exceptions.UsageError if parsing failed!
    #      However, exception handling cannot happen here because the new
    #      context would not be usable. Thus, we let these propagate upwards.
    new_ctx = old_ctx.command.make_context(
        "utopya eval", args=copy.copy(new_args)
    )
    new_params = new_ctx.params

    # Check that bad arguments were not used
    defaults = {p.name: p.get_default(new_ctx) for p in new_ctx.command.params}
    bad_args = {
        arg: new_params[arg]
        for arg in INTERACTIVE_MODE_PROHIBITED_ARGS
        if new_params.get(arg) != defaults.get(arg)
    }
    if bad_args:
        raise click.exceptions.UsageError(
            "During interactive plotting, arguments that update the "
            "Multiverse meta-configuration cannot be used!\n"
            "Remove the offending argument{} (value shown in parentheses):\n{}"
            "".format(
                "s" if len(bad_args) != 1 else "",
                "\n".join({f"  {k}  ({v})" for k, v in bad_args.items()}),
            )
        )

    return new_args, new_params, new_ctx



"""Utility functions that work on the already initialized model registry."""

import logging
from typing import Tuple, Union

from .._yaml import load_yml
from ..parameter import extract_validation_objects
from . import MODELS
from .info_bundle import ModelInfoBundle

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


def get_info_bundle(
    *,
    model_name: str = None,
    info_bundle: ModelInfoBundle = None,
    bundle_label: Union[str, int] = None,
) -> ModelInfoBundle:
    """Determine the model info bundle in cases where both a model name
    and an info bundle are allowed as arguments.

    Args:
        model_name (str, optional): The model name.
        info_bundle (ModelInfoBundle, optional): The info bundle object.
            If given, will directly return this object again.
        bundle_label (Union[str, int], optional): In cases where only the model
            name is given, the bundle_label can be used for item access,
            e.g. in cases where more than one bundle is available and access
            would be ambiguous.

    Returns:
        ModelInfoBundle: The selected info bundle item

    Raises:
        ValueError: If neither or both model_name and info_bundle were None
    """
    if model_name is None == info_bundle is None:  # XOR-check existence
        raise ValueError(
            "Need either of the arguments model_name or "
            "info_bundle, but got both or neither!"
        )

    if info_bundle:
        return info_bundle

    if bundle_label is None:
        return MODELS[model_name].item()
    return MODELS[model_name][bundle_label]


def load_model_cfg(**get_info_bundle_kwargs) -> Tuple[dict, str, dict]:
    """Loads the default model configuration file for the given model name,
    using the path specified in the info bundle.

    Furthermore, :py:func:`~utopya.parameter.extract_validation_objects` is
    called to extract any Parameter objects that require validation, replace
    them with their default values, and gather the Parameter class objects
    into a separate dict.

    Args:
        **get_info_bundle_kwargs: Used to retrieve the info bundle via
            :py:func:`~utopya.model_registry.utils.get_info_bundle`

    Returns:
        Tuple[dict, str, dict]: The corresponding model configuration, the path
            to the model configuration file, and the Parameter class objects
            requiring validation.
            Will be ``{}, None, {}`` if there is no default configuration
            available in the selected info bundle.

    Raises:
        FileNotFoundError: On missing file
    """
    bundle = get_info_bundle(**get_info_bundle_kwargs)
    if "default_cfg" not in bundle.paths:
        log.debug(
            "No default model configuration available for '%s'.",
            bundle.model_name,
        )
        return {}, None, {}

    path = bundle.paths["default_cfg"]

    log.debug(
        "Loading default model configuration for '%s' model ...\n  %s",
        bundle.model_name,
        path,
    )

    try:
        model_cfg = load_yml(path)

    except FileNotFoundError as err:
        raise FileNotFoundError(
            "Could not locate default configuration for "
            f"'{bundle.model_name}' model! Expected to find it at: {path}"
        ) from err

    # Collect the validation objects from the model configuration and replace
    # them with their default values
    model_cfg, to_validate = extract_validation_objects(
        model_cfg, model_name=bundle.model_name
    )
    return model_cfg, path, to_validate



"""Implementation of the utopya project registry"""

import copy
import logging
import os
from typing import Any, Dict, List, Optional, Sequence

from pydantic import DirectoryPath, FilePath

from ._yaml_registry import BaseSchema, RegistryEntry, YAMLRegistry
from .cfg import PROJECT_INFO_FILE_SEARCH_PATHS, UTOPYA_CFG_SUBDIRS
from .tools import load_yml, pformat, recursive_update

log = logging.getLogger(__name__)

# -- Schema definition --------------------------------------------------------
# .. Sub-schemas ..............................................................


class ProjectPaths(BaseSchema):
    """Schema to use for a project's ``paths`` field"""

    base_dir: DirectoryPath
    project_info: Optional[FilePath]
    models_dir: Optional[DirectoryPath]
    py_tests_dir: Optional[DirectoryPath]
    py_plots_dir: Optional[DirectoryPath]
    mv_project_cfg: Optional[FilePath]
    project_base_plots: Optional[FilePath]

    # TODO What about the BatchTaskManager configuration?


class ProjectMetadata(BaseSchema):
    """Schema to use for a project's ``metadata`` field"""

    version: Optional[str]
    long_name: Optional[str]
    description: Optional[str]
    long_description: Optional[str]
    license: Optional[str]
    authors: Optional[List[str]]
    email: Optional[str]
    website: Optional[str]
    utopya_compatibility: Optional[str]
    language: Optional[str]
    requirements: Optional[List[str]]
    misc: Optional[Dict[str, Any]]


# .............................................................................


class ProjectSchema(BaseSchema):
    """The data model for a project registry entry"""

    project_name: str
    framework_name: Optional[str]
    paths: ProjectPaths
    metadata: ProjectMetadata
    run_cfg_format: str = "yaml"
    cfg_set_abs_search_dirs: Optional[List[str]]
    cfg_set_model_source_subdirs: Optional[List[str]]
    custom_py_modules: Optional[Dict[str, DirectoryPath]]
    output_files: Optional[dict]  # TODO Needs sub-schema
    debug_level_updates: Optional[Dict[str, dict]]  # TODO Implement


# -- Project ------------------------------------------------------------------


class Project(RegistryEntry):
    """A registry entry that describes a project"""

    SCHEMA = ProjectSchema

    @property
    def framework_project(self) -> Optional["Project"]:
        """If a framework project is defined, retrieve it from the registry"""
        if not self.framework_name:
            return None

        from . import PROJECTS

        return PROJECTS[self.framework_name]


# -- ProjectRegistry ----------------------------------------------------------


class ProjectRegistry(YAMLRegistry):
    """The project registry"""

    def __init__(self):
        """Initializes the project registry, loading available entries from
        the registry directory in the utopya config directory.

        This also creates the ``projects`` directory, if not created yet.
        """

        registry_dir = UTOPYA_CFG_SUBDIRS["projects"]
        if not os.path.exists(registry_dir):
            os.makedirs(registry_dir)

        super().__init__(Project, registry_dir=registry_dir)

    def register(
        self,
        *,
        base_dir: str,
        info_file: str = None,
        custom_project_name: str = None,
        require_matching_names: bool = None,
        exists_action: str = "raise",
    ) -> dict:
        """Register or update information of a project.

        Args:
            base_dir (str): Project base directory
            info_file (str, optional): Path to info file which contains further
                path information and metadata (may be relative to base
                directory). If not given, will use some defaults to search for
                it.
            custom_project_name (str, optional): Custom project name,
                overwrites the one given in the info file
            require_matching_names (bool, optional): If set, will require that
                the custom project name is equal to the one given in the
                project info file. This allows checking that the file content
                does not diverge from some outside state.
            exists_action (str, optional): Action to take upon existing project

        Returns:
            dict: Project information for the new or validated project
        """
        log.info("Commencing project registration ...")

        # Parse base directory and info file; if not given, search for it
        base_dir = os.path.abspath(os.path.expanduser(base_dir))
        log.remark("Project base directory:\n  %s", base_dir)

        if info_file:
            info_file = os.path.realpath(
                os.path.join(base_dir, os.path.expanduser(info_file))
            )
            log.remark("Have project info file available:\n  %s", info_file)

        else:
            for _relpath in PROJECT_INFO_FILE_SEARCH_PATHS:
                _abspath = os.path.join(base_dir, _relpath)
                if os.path.isfile(_abspath):
                    info_file = _abspath
                    log.remark("Found project info file:\n  %s", info_file)
                    break

            else:
                _locs = "\n".join(
                    f"  - {p}" for p in PROJECT_INFO_FILE_SEARCH_PATHS
                )
                raise ValueError(
                    "Missing project info file!\nEither explicitly supply a "
                    "path or add one at the following search locations "
                    f"relative to the base directory:\n{_locs}"
                )

        # Load and parse the info file
        log.note("Loading project information from file ...")

        project_info = dict(paths=dict(base_dir=base_dir))
        _project_info = load_yml(info_file)
        _project_info = _project_info if _project_info else {}

        # Warn if base_dir was given in info file; the base directory given
        # explicitly should always take precedence
        if _project_info.get("paths", {}).get("base_dir"):
            log.caution(
                "The project info file contains a `paths.base_dir` entry, "
                "which will be ignored!"
            )

        project_info = recursive_update(_project_info, project_info)
        project_info["paths"]["project_info"] = info_file

        # Regularize paths, evaluating relative ones towards the base directory
        for path_name, path in project_info["paths"].items():
            if path_name in ("base_dir", "project_info"):
                continue

            project_info["paths"][path_name] = os.path.realpath(
                os.path.join(base_dir, path)
            )

        # May want to use a custom project name
        if custom_project_name:
            _project_name = project_info.get("project_name")
            if require_matching_names and custom_project_name != _project_name:
                raise ValueError(
                    "The custom project name '{}' does not match the name "
                    "given in the project info file, '{}'! "
                    "Either ensure that the names match exactly or unset the "
                    "`require_matching_names` flag.".format(
                        custom_project_name, project_info["project_name"]
                    )
                )

            project_info["project_name"] = custom_project_name
            log.remark("Using a custom project name:  %s", custom_project_name)
            log.remark(
                "Note that this creates a mismatch between the project "
                "info file and the registered project!"
            )

        # Now, create a registry entry from it
        log.note(
            "Preparing to add entry (exists_action: %s) ...", exists_action
        )
        project_name = project_info["project_name"]

        project = self.add_entry(
            project_name, exists_action=exists_action, **project_info
        )

        log.success(
            "Successfully stored or updated '%s' project information.",
            project_name,
        )
        log.remark(
            "NOTE: Changes to the project info file are *not* automatically "
            "tracked!\nTo update project information, repeat this procedure."
        )

        return self[project_name]


# -----------------------------------------------------------------------------

PROJECTS = ProjectRegistry()
"""The package-wide project registry"""



"""The plot_funcs subpackage supplies general plotting functions.

These can be used via the plotting framework and its ExternalPlotCreator, of
which this subpackage is the base package to use for relative module imports.

It extends those capabilities provided by the dantro plotting framework.
"""

from dantro.plot import is_plot_func

from ..plotcreators import MultiversePlotCreator, UniversePlotCreator
from ..plothelper import PlotHelper



"""Overloads and additions for dantro's data transformation framework."""

from typing import Callable, Union

import dantro.data_ops as dops

# -----------------------------------------------------------------------------


def register_operation(*args, skip_existing: bool = True, **kws) -> None:
    """Register an operation with the dantro data operations database.

    This invokes :py:func:`~dantro.data_ops.db_tools.register_operation`, but
    has ``skip_existing == True`` as default in order to reduce number of
    arguments that need to be specified in utopya model plots, where duplicate
    module imports frequently cause existing entries.

    Args:
        *args: Passed to :py:func:`dantro.data_ops.db_tools.register_operation`
        skip_existing (bool, optional): Whether to skip (without an error) if
            an operation
        **kws: Passed to :py:func:`dantro.data_ops.db_tools.register_operation`
    """
    return dops.register_operation(*args, **kws, skip_existing=skip_existing)


def is_operation(arg: Union[str, Callable] = None, /, **kws):
    """Overload of dantro's ``is_operation`` decorator, using utopya's own
    registration function.

    Usage example:

    .. testcode:: is_operation_decorator

        from utopya.eval import is_operation

        @is_operation
        def my_operation(data, *args):
            pass

        @is_operation("op_with_custom_name")
        def my_other_operation(data, *args):
            pass

        @is_operation("my_operation", overwrite_existing=True)
        def yet_some_other_operation(data, *args):
            pass

    .. testcode:: is_operation_decorator
        :hide:

        from dantro.data_ops.db import _OPERATIONS
        assert "my_operation" in _OPERATIONS
        assert "op_with_custom_name" in _OPERATIONS

    Args:
        arg (Union[str, Callable], optional): The name that should be used in
            the operation registry. If not given, will use the name of the
            decorated function instead. If a callable, this refers to the
            ``@is_operation`` call syntax and will use that as a function.
        **kws: Passed to :py:func:`~.register_operation`.
    """
    return dops.is_operation(arg, _reg_func=register_operation, **kws)



"""Sets up the utopya package, test dependencies, and command line scripts"""

from setuptools import find_packages, setup

# .. Dependency lists .........................................................

INSTALL_DEPS = [
    "numpy>=1.21",
    "scipy>=1.7.3",
    "matplotlib>=3.3",
    "seaborn>=0.11.2",
    "ruamel.yaml>=0.16.5",
    "coloredlogs>=15.0",
    "colorama>=0.4.4",
    "click>=8.0",
    "pydantic>=1.9",
    #
    # related to utopya:
    "paramspace>=2.5.9",
    "dantro>=0.18.0a11",
]
# NOTE When adding a new dependency, make sure to denote it in the isort
#      configuration, see pyproject.toml.

# Dependencies for running tests and general development of dantro
TEST_DEPS = [
    "pytest>=6.2",
    "pytest-cov>=2.10",
    "pre-commit>=2.16",
]

# Dependencies for building the dantro documentation
DOC_DEPS = [
    "sphinx>=4.5,<5",
    "sphinx-book-theme",
    "ipython>=7.0",
]


# .. Description ..............................................................

DESCRIPTION = "A simulation management and evaluation framework"
LONG_DESCRIPTION = """
``utopya``: A simulation management and evaluation framework
============================================================

The ``utopya`` package provides a simulation management and evaluation
framework with the following feature set:

- **Run model simulations** in parallel and on cluster architectures

  - Conveniently perform parameter sweeps of arbitrary parameters with the help
    of the `paramspace <https://gitlab.com/blsqr/paramspace>`_ package.

- A **powerful CLI** to run and evaluate models, including interactive plotting
- Integrates the `dantro <https://gitlab.com/utopia-project/dantro>`_
  **data processing pipeline**:

  - Loads data into a hierarchical data tree, supplying a uniform interface
  - Gives access to a configuration-based **data transformation framework**,
    separating data preprocessing from visualization for increased generality
  - Easy extensibility of plot creators via model-specific plot implementations

- A **versatile configuration interface** for both simulation and evaluation:

  - Assembling multi-level model configurations, including several default
    levels
  - Assembling plot configurations with multiple inheritance, reducing
    redundant definitions

- Model, project, and framework registration and handling
- Managing data output directories
- Tools to simplify model test implementations or working without a CLI
- ... and more

The ``utopya`` package evolved as part of the
`Utopia Project <https://utopia-project.org>`_ and provides the frontend of
the `Utopia modelling framework <https://gitlab.com/utopia-project/utopia>`_.
Having been outsourced from that project, it can be used with arbitrary model
implementations with a very low barrier for entry: in the simplest case, only
the path to an executable is required to run simulations.
With more compliance to the utopya interface, more features become available.

The ``utopya`` package is **open source software** released under the
`LGPLv3+ <https://www.gnu.org/licenses/lgpl-3.0.html>`_ license.

Visit the `utopya project website <https://gitlab.com/utopia-project/utopya>`_
for more information and documentation of this package.
"""
# TODO Mention docs


# .............................................................................


def find_version(*file_paths) -> str:
    """Tries to extract a version from the given path sequence"""
    import codecs
    import os
    import re

    def read(*parts):
        """Reads a file from the given path sequence, relative to this file"""
        here = os.path.abspath(os.path.dirname(__file__))
        with codecs.open(os.path.join(here, *parts), "r") as fp:
            return fp.read()

    # Read the file and match the __version__ string
    file = read(*file_paths)
    match = re.search(r"^__version__\s?=\s?['\"]([^'\"]*)['\"]", file, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Unable to find version string in " + str(file_paths))


# .............................................................................


setup(
    name="utopya",
    #
    # Package information
    version=find_version("utopya", "__init__.py"),
    #
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url="https://gitlab.com/utopia-project/utopya",
    author="utopya developers",
    author_email=(
        "Benjamin Herdeanu <Benjamin.Herdeanu@iup.uni-heidelberg.de>, "
        "Yunus Sevinchan <Yunus.Sevinchan@iup.uni-heidelberg.de>"
    ),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Topic :: Utilities",
    ],
    #
    # Package content and dependencies
    packages=find_packages(exclude=("demo", "tests")),
    package_data=dict(utopya=["cfg/*.yml"]),
    install_requires=INSTALL_DEPS,
    extras_require=dict(
        test=TEST_DEPS,
        doc=DOC_DEPS,
        dev=TEST_DEPS + DOC_DEPS,
    ),
    #
    # Command line scripts
    entry_points={
        "console_scripts": [
            "utopya = utopya_cli.cli:cli",
        ],
    },
)



"""Supplies basic YAML interface, inherited from dantro and paramspace"""

import paramspace
from dantro.tools import load_yml, write_yml, yaml

# Make sure the yaml instances are really shared
assert yaml is paramspace.yaml



"""Implements a class that manages data written out by Utopia models.

It is based on dantro's ``DataManager`` class and the containers specialized
for Utopia data.
"""

import logging

import dantro
import dantro.data_loaders
import dantro.data_mngr

from .containers import GridDC, NumpyDC, XarrayDC, XarrayYamlDC
from .groups import (
    GraphGroup,
    HeterogeneousTimeSeriesGroup,
    MultiverseGroup,
    TimeSeriesGroup,
    UniverseGroup,
)

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


def _condense_thresh_func(
    *, level: int, num_items: int, total_item_count: int
) -> int:
    """Dynamically computes the condensation threshold for the current
    element in a :py:class:`~utopya.eval.datamanager.DataManager` tree.
    """
    # For high item counts, always condense
    if total_item_count > 100:  # NOTE This is along *one* recursion branch!
        return 7  # shows first three and last three

    # Now, distinguish by level
    if level == 1:
        # Level of the universes 0, ..., N
        if num_items > 3:
            return 3

    elif level >= 4:
        # Level of the model data, e.g. multiverse/0/data/dummy/...
        if num_items > 15:
            return 5

    # All other cases: do not condense
    return None


# -----------------------------------------------------------------------------


class DataManager(
    dantro.data_loaders.AllAvailableLoadersMixin, dantro.data_mngr.DataManager
):
    """This class manages the data that is written out by Utopia simulations.

    It is based on the dantro ``DataManager`` class and adds the functionality
    for specific loader functions that are needed in Utopia: Hdf5 and Yaml.
    Furthermore, to enable file caching via the DAG framework, all available
    data loaders are included here.
    """

    _DATA_GROUP_CLASSES = dict(
        MultiverseGroup=MultiverseGroup,
        GraphGroup=GraphGroup,
    )
    """Known group types"""

    _HDF5_DSET_DEFAULT_CLS = XarrayDC
    """Tells the HDF5 loader which container class to use"""

    _HDF5_MAP_FROM_ATTR = "content"
    """The name of the attribute to read for the mapping"""

    _HDF5_GROUP_MAP = dict(
        network=GraphGroup,
        graph=GraphGroup,
        time_series=TimeSeriesGroup,
        time_series_heterogeneous=HeterogeneousTimeSeriesGroup,
    )
    """The mapping of different content values to a data group type"""

    _HDF5_DSET_MAP = dict(
        grid=GridDC,
        unlabelled_data=NumpyDC,
        labelled_data=XarrayDC,
        array_of_yaml_strings=XarrayYamlDC,
    )
    """The mapping of different content values to a data container types"""

    _COND_TREE_MAX_LEVEL = 10
    """Condensed tree representation maximum level"""

    _COND_TREE_CONDENSE_THRESH = lambda _, **kws: _condense_thresh_func(**kws)
    """Condensed tree representation threshold parameter"""

    _DEFAULT_TREE_CACHE_PATH = "data/.tree_cache.d3"
    """Where the tree cache file is to be stored; may be overwritten by a
    config entry
    """



"""Implements the `utopya batch` subcommand"""

import click


@click.command(
    help=(
        "Run and evaluate multiple simulations.\n"
        "\n"
        "Which batch tasks are to be performed is determined using a "
        "batch configuration file."
    ),
)
@click.argument(
    "batch_cfg_path",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
)
@click.option(
    "-d",
    "--debug",
    is_flag=True,
    default=False,
    help="If set, an error in an individual task will stop all other tasks.",
)
@click.option(
    "-s",
    "--single-worker",
    is_flag=True,
    default=False,
    help="If set, will use 'task' parallelization level instead of 'batch'. "
    "Effectively, the batch tasks will be worked on sequentially. "
    "This option can also be useful for debugging or when encountering "
    "memory issues.",
)
@click.option(
    "--note",
    default=None,
    type=str,
    help=(
        "Overwrites the `paths->model_note` entry which is used in creation "
        "of the batch run directory path."
    ),
)
def batch(
    batch_cfg_path: str,
    debug: bool,
    single_worker: bool,
    note: str,
):
    from utopya.batch import BatchTaskManager

    kws = dict()
    if debug:
        kws["debug"] = debug

    if single_worker:
        kws["parallelization_level"] = "task"

    if note:
        kws["paths"] = dict(note=note)

    btm = BatchTaskManager(batch_cfg_path=batch_cfg_path, **kws)
    btm.perform_tasks()

    Echo.success("Batch work all finished now, yay! :)\n")
    sys.exit()



"""Takes care of the YAML setup for Utopya.

In the module import order, this module needs to be downstream from all modules
that implement objects that require a custom YAML representation."""

import functools
import logging
import re
from typing import Callable, Tuple

import numpy as np
import paramspace.yaml_constructors as pspyc
import ruamel.yaml

from ._yaml import load_yml, write_yml, yaml
from .model_registry import ModelInfoBundle, ModelRegistryEntry, load_model_cfg
from .parameter import Parameter
from .stop_conditions import StopCondition
from .tools import recursive_update as _recursive_update

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


def _scalar_node_to_object(loader, node):
    """Attempts to convert the given scalar node to a null (Python None),
    a bool, an int, or a float object using the corresponding YAML constructor.
    If those conversions fail, constructs a scalar (which will typically result
    in a string being returned).
    """

    def construct_yaml_null(node) -> None:
        """Constructs a None from an appropriate YAML node.

        This custom constructor should not be necessary, but for some weird
        reason, the ruamel.yaml constructor from loader.construct_yaml_null
        *always* returns None, regardless of the value of the node. As we rely
        on errors being raised if construction fails, we need this custom
        constructor for the two explicitly allowed null values.
        """
        if node.value in ("~", "null"):
            return None
        raise ruamel.yaml.constructor.ConstructorError(
            f"expected null, but " f"got '{node.value}'"
        )

    for constructor in (
        loader.construct_yaml_bool,
        loader.construct_yaml_int,
        loader.construct_yaml_float,
        construct_yaml_null,
        loader.construct_yaml_str,
    ):
        try:
            return constructor(node)
        except:
            pass

    return loader.construct_scalar(node)


# -----------------------------------------------------------------------------


def _expr_constructor(loader, node):
    """Custom pyyaml constructor for evaluating strings with simple
    mathematical expressions.

    Supports: +, -, *, **, /, e-X, eX, inf, nan
    """
    # get expression string
    expr_str = loader.construct_scalar(node)

    # Remove spaces
    expr_str = expr_str.replace(" ", "")

    # Parse some special strings
    if expr_str in ["nan", "NaN"]:
        return float("nan")

    # NOTE these will cause errors if emitted file is not read by python!
    elif expr_str in ["np.inf", "inf", "INF"]:
        return np.inf

    elif expr_str in ["-np.inf", "-inf", "-INF"]:
        return -np.inf

    # remove everything that might cause trouble -- only allow digits, dot, +,
    # -, *, /, and eE to allow for writing exponentials, and parentheses
    expr_str = re.sub(r"[^0-9eE\-.+\*\/\(\)]", "", expr_str)

    # Try to eval
    return eval(expr_str)


def _model_cfg_constructor(loader, node) -> dict:
    """Custom yaml constructor for loading a model configuration file.

    This extracts the `model_name` key, loads the corresponding model config
    and then recursively updates the loaded config with the remaining keys
    from that part of the configuration.
    """
    # Get a mapping from the node
    d = loader.construct_mapping(node, deep=True)
    # NOTE using the deep flag here to allow nested calls to this constructor

    # Extract the model name and a potentially existing bundle key
    model_name = d.pop("model_name")
    bundle_label = d.pop("bundle_label", None)

    # Load the corresponding model configuration
    mcfg, _, _ = load_model_cfg(
        model_name=model_name, bundle_label=bundle_label
    )

    # Update the loaded config with the remaining keys
    mcfg = _recursive_update(mcfg, d)

    # Return the updated dictionary
    return mcfg


def _func_on_sequence_constructor(loader, node, *, func: Callable):
    """Custom yaml constructor that constructs a sequence, passes it to the
    given function, and returns the result of that call.

    Can be used e.g. in conjunction with the any and all functions, evaluating
    sequences of booleans.
    """
    # Get a sequence from the node
    s = loader.construct_sequence(node, deep=True)
    return func(s)


def _parameter_shorthand_constructor(loader, node) -> Parameter:
    """Constructs a Parameter object from a scalar YAML node using
    :py:func:`~utopya.yaml._scalar_node_to_object`.

    The YAML tag is used as shorthand ``mode`` argument to the
    :py:class:`~utopya.parameter.Parameter.from_shorthand` class method.
    """
    return Parameter.from_shorthand(
        _scalar_node_to_object(loader, node), mode=node.tag[1:]
    )


# -----------------------------------------------------------------------------
# Attaching representers and constructors

# First register the classes which directly implemented dumping/loading
yaml.register_class(StopCondition)
yaml.register_class(ModelInfoBundle)
yaml.register_class(ModelRegistryEntry)
yaml.register_class(Parameter)

# Now, add (additional, potentially overwriting) constructors for certain tags.
# Evaluate a mathematical expression
yaml.constructor.add_constructor("!expr", _expr_constructor)

# Apply the any operator to a sequence
yaml.constructor.add_constructor(
    "!any", functools.partial(_func_on_sequence_constructor, func=any)
)

# Apply the all operator to a sequence
yaml.constructor.add_constructor(
    "!all", functools.partial(_func_on_sequence_constructor, func=all)
)

# Load a model configuration
yaml.constructor.add_constructor("!model", _model_cfg_constructor)


# Add aliases for the (coupled) parameter dimensions
yaml.constructor.add_constructor("!sweep", pspyc.pdim)
yaml.constructor.add_constructor("!sweep-default", pspyc.pdim_default)

yaml.constructor.add_constructor("!coupled-sweep", pspyc.coupled_pdim)
yaml.constructor.add_constructor(
    "!coupled-sweep-default", pspyc.coupled_pdim_default
)


# Register the Parameter shorthand constructors
# NOTE The regular constructors and representers are already registered using
#      the call to yaml.register_class above
for mode in Parameter.SHORTHAND_MODES:
    yaml.constructor.add_constructor(
        "!" + mode, _parameter_shorthand_constructor
    )

# Set the flow style
yaml.default_flow_style = False



"""Implements functionality to copy relevant model files"""

import glob
import os
from typing import Callable, Dict, List, Sequence, Tuple

FILE_EXTENSIONS = {
    "c": (
        ".c",
        ".h",
    ),
    "c++": (".c", ".h", ".cc", ".hh", ".cpp"),
    "python": (".py", ".pyx"),
    "yaml": (".yml", ".yaml"),
}
"""A map of language-specific file extensions"""

LANGUAGE_ALIASES = {
    "cpp": "c++",
    "py": "python",
    "py2": "python",
    "py3": "python",
}
"""A map of language specifier aliases that can be used to find a normalized
language specifier"""

# -----------------------------------------------------------------------------


def copy_model_files(
    *,
    model_name: str,
    new_name: str = None,
    target_project: str = None,
    add_to_cmakelists: bool = True,
    skip_exts: Sequence[str] = None,
    use_prompts: bool = True,
    dry_run: bool = False,
) -> None:
    """A helper function to conveniently copy model-related files, rename them,
    and adjust their content to the new name as well.

    Args:
        model_name (str): The name of the model to copy
        new_name (str, optional): The new name of the model. This may not
            conflict with any already existing model name in the model
            registry.
        target_project (str, optional): The name of the project to copy the
            model to. It needs to be a registered Utopia project.
        add_to_cmakelists (bool, optional): Whether to add the new model to the
            corresponding CMakeLists.txt file.
        use_prompts (bool, optional): Whether to interactively prompt for
            confirmation or missing arguments.
        dry_run (bool, optional): If given, no write or copy operations will be
            carried out.

    Raises:
        ValueError: Upon bad arguments

    Returns:
        None
    """
    # FIXME Only import needed ones!
    from utopya import MODELS as _MODELS
    from utopya.cfg import load_from_cfg_dir, write_to_cfg_dir
    from utopya.model_registry import get_info_bundle as _get_info_bundle
    from utopya.multiverse import Multiverse
    from utopya.tools import add_item, recursive_update

    def apply_replacements(s, *replacements: Sequence[Tuple[str, str]]) -> str:
        """Applies multiple replacements onto the given string"""
        for replacement in replacements:
            s = s.replace(*replacement)
        return s

    def create_file_map(
        *,
        source_dir: str,
        target_dir: str,
        abs_file_map: dict,
        replacements: Sequence[Tuple[str, str]],
        skip_exts: Sequence[str] = None,
        glob_args: Sequence[str] = ("**",),
    ) -> dict:
        """Given a file list with absolute paths, aggregates the file path
        changes into ``abs_file_map`` and gathers the relative file path
        changes into the returned dict.

        The file name is changed according to the specified replacements.

        Args:
            source_dir (str): The source directory to look for files in using
                glob and the ``glob_args``. Note that directories are not
                matched.
            target_dir (str): The target directory of the renamed files
            abs_file_map (dict): The mutable file map that the absolute file
                path changes are aggregated in.
            replacements (Sequence[Tuple[str, str]]): The replacement
                specifications, applied to the relative paths.
            glob_args (Sequence[str], optional): The glob arguments to match
                files within the source directory. By default, this matches
                all files, also down the source directory tree. The glob
                ``recursive`` option is enabled.

        Returns:
            dict: The file map relative to source and target dir.
        """
        files = glob.glob(os.path.join(source_dir, *glob_args), recursive=True)
        rel_file_map = dict()

        for fpath in files:
            if os.path.isdir(fpath) or not os.path.exists(fpath):
                continue

            if skip_exts and os.path.splitext(fpath)[1] in skip_exts:
                continue

            rel_fpath = os.path.relpath(fpath, start=source_dir)
            new_rel_fpath = apply_replacements(rel_fpath, *replacements)

            abs_file_map[fpath] = os.path.join(target_dir, new_rel_fpath)
            rel_file_map[rel_fpath] = new_rel_fpath

        return rel_file_map

    def print_file_map(
        *, file_map: dict, source_dir: str, target_dir: str, label: str
    ):
        """Prints a human-readable version of the given (relative) file map
        which copies from the source directory tree to the target directory
        tree.
        """
        max_key_len = min(max(len(k) for k in file_map), 32)
        files = [
            "\t{:{l:d}s}  ->  {:s}".format(k, v, l=max_key_len)
            for k, v in file_map.items()
        ]

        print(
            "\nThe following {num:d} {label:s} files from\n\t{from_dir:}\n"
            "will be copied to\n\t{to_dir:}\nusing the following new file "
            "names:\n{files:}"
            "".format(
                num=len(file_map),
                label=label,
                from_dir=source_dir,
                to_dir=target_dir,
                files="\n".join(files),
            )
        )

    def add_model_to_cmakelists(*, fpath: str, new_name: str, write: bool):
        """Adds the relevant add_subdirectory command to the CMakeLists file
        at the specified path.

        Assumes an ascending alphabetical list of add_subdirectory commands
        and adds the new command at a suitable place.

        Args:
            fpath (str): The absolute path of the CMakeLists.txt file
            new_name (str): The new model name to add to it
            write (bool): If false, will not write.

        Raises:
            ValueError: On missing ``add_subdirectory`` command in the given
                file. In this case, the line has to be added manually.
        """
        # Read the file
        with open(fpath) as f:
            lines = f.readlines()

        # Find the line to add the add_subdirectory command at
        insert_idx = None
        for i, line in enumerate(lines):
            if not line.startswith("add_subdirectory"):
                continue

            insert_idx = i
            _model = line[len("add_subdirectory(") : -2]
            if _model.lower() > new_name.lower():
                break
        else:
            # Did not break. Insert behind the last add_subdirectory command
            insert_idx += 1

        if insert_idx is None:
            raise ValueError(
                "Found no add_subdirectory commands and thus do "
                "not know where to insert the command for the "
                "new model directory; please do it manually in "
                f"the following file:  {fpath}"
            )

        lines.insert(
            insert_idx if insert_idx else last_add_subdir_idx + 1,
            f"add_subdirectory({new_name})\n",
        )

        if write:
            with open(fpath, "w") as f:
                f.writelines(lines)

            print(f"Subdirectory for model '{new_name}' added to\n\t{fpath}")

        else:
            print(
                f"Not writing. Preview of how the new\n\t{fpath}\n"
                "file _would_ look like:"
            )
            print("-" * 79 + "\n")
            print("".join(lines))
            print("-" * 79)

    # Gather information on model, project, and replacements . . . . . . . . .
    # Get the model information
    info_bundle = _get_info_bundle(model_name=model_name)
    print(
        f"\nModel selected to copy:     {info_bundle.model_name}  "
        f"(from project: {info_bundle.project_name})"
    )

    # Find out the new name
    if not new_name:
        if not use_prompts:
            raise ValueError("Missing new_name argument!")
        try:
            new_name = input("\nWhat should be the name of the NEW model?  ")
        except KeyboardInterrupt:
            return

    # Check if the name is not already taken, being case-insensitive
    if new_name.lower() in [n.lower() for n in _MODELS.keys()]:
        _avail = ", ".join(_MODELS.keys())
        raise ValueError(
            f"A model with name '{new_name}' is already registered! "
            "Make sure that the name is unique. If you keep "
            "receiving this error despite no other model with "
            "this name being implemented, remove the entry from "
            "the model registry, e.g. via the `utopia models rm` "
            "CLI command.\n"
            f"Already registered models: {_avail}"
        )

    print(f"Name of the new model:      {new_name}")

    # Define the replacements
    replacements = [
        (model_name, new_name),
        (model_name.lower(), new_name.lower()),
        (model_name.upper(), new_name.upper()),
    ]

    # Find out the project that the files are copied _to_
    projects = load_from_cfg_dir("projects")

    if not target_project:
        if not use_prompts:
            raise ValueError("Missing target_project argument!")
        try:
            target_project = input(
                f"\nWhich Utopia project (available: {', '.join(projects)}) "
                "should the model be copied to?  "
            )
        except KeyboardInterrupt:
            return
    print(f"Utopia project to copy to:  {target_project}")

    project_info = projects.get(target_project)
    if not project_info:
        raise ValueError(
            f"No Utopia project with name '{target_project}' is known to the "
            "frontend. Check the spelling and note that the "
            "project name is case-sensitive.\n"
            f"Available projects: {', '.join(projects)}."
        )

    # Generate the file maps . . . . . . . . . . . . . . . . . . . . . . . . .
    # The mapping of all files that are to be copied and in which the content
    # is to be replaced. It maps absolute source file paths to absolute target
    # file paths.
    file_map = dict()

    # Relative file maps, created below
    impl_file_map = None
    py_t_file_map = None
    py_p_file_map = None

    # Find out the target directories
    target_models_dir = project_info.get("models_dir")
    target_py_t_dir = project_info.get("python_model_tests_dir")
    target_py_p_dir = project_info.get("python_model_plots_dir")

    # Define the source and target directory paths of the implementation and
    # the python-related files, if the path information is available.
    impl_source_dir = info_bundle.paths["source_dir"]
    impl_target_dir = os.path.join(target_models_dir, new_name)
    impl_file_map = create_file_map(
        source_dir=impl_source_dir,
        target_dir=impl_target_dir,
        abs_file_map=file_map,
        replacements=replacements,
        skip_exts=skip_exts,
    )

    if target_py_t_dir and info_bundle.paths.get("python_model_tests_dir"):
        py_t_source_dir = info_bundle.paths["python_model_tests_dir"]
        py_t_target_dir = os.path.join(target_py_t_dir, new_name)
        py_t_file_map = create_file_map(
            source_dir=py_t_source_dir,
            target_dir=py_t_target_dir,
            abs_file_map=file_map,
            replacements=replacements,
            skip_exts=skip_exts,
        )

    if target_py_p_dir and info_bundle.paths.get("python_model_plots_dir"):
        py_p_source_dir = info_bundle.paths["python_model_plots_dir"]
        py_p_target_dir = os.path.join(target_py_p_dir, new_name)
        py_p_file_map = create_file_map(
            source_dir=py_p_source_dir,
            target_dir=py_p_target_dir,
            abs_file_map=file_map,
            replacements=replacements,
            skip_exts=skip_exts,
        )

    # Gathered all information now. . . . . . . . . . . . . . . . . . . . . . .
    # Inform about the file changes and the replacement in them.
    print_file_map(
        file_map=impl_file_map,
        label="model implementation",
        source_dir=impl_source_dir,
        target_dir=impl_target_dir,
    )

    if py_t_file_map:
        print_file_map(
            file_map=py_t_file_map,
            label="python model test",
            source_dir=py_t_source_dir,
            target_dir=py_t_target_dir,
        )

    if py_p_file_map:
        print_file_map(
            file_map=py_p_file_map,
            label="python model plot",
            source_dir=py_p_source_dir,
            target_dir=py_p_target_dir,
        )

    max_repl_len = max(len(rs) for rs, _ in replacements)
    repl_info = [
        "\t'{:{l:d}s}'  ->  '{:s}'".format(*repl, l=max_repl_len)
        for repl in replacements
    ]
    _repl_info = "\n".join(repl_info)
    print(
        f"\nInside all of these {len(file_map):d} files, the following string "
        f"replacements will be carried out:\n{_repl_info}\n"
    )

    # Inform about dry run and ask whether to proceed
    if dry_run:
        print("--- THIS IS A DRY RUN. ---")
        print("Copy and write operations below are not operational.")

    if use_prompts:
        try:
            response = input("\nProceed [y/N]  ")
        except KeyboardInterrupt:
            response = "N"
        if response.lower() not in ("y", "yes"):
            print("\nNot proceeding ...")
            return

    print("\nNow copying and refactoring ...")

    # Now, the actual copying . . . . . . . . . . . . . . . . . . . . . . . . .
    for i, (src_fpath, target_fpath) in enumerate(file_map.items()):
        print(f"\nFile {i + 1:d}/{len(file_map):d} ...")
        print(f"\t   {src_fpath:s}\n\t-> {target_fpath:s}")

        try:
            with open(src_fpath) as src_file:
                src_lines = src_file.read()

        except Exception as exc:
            print(f"\tReading FAILED due to {type(exc).__name__}: {str(exc)}.")
            print(
                "\tIf you want this file copied and refactored, you will "
                "have to do it manually."
            )
            continue

        target_lines = apply_replacements(src_lines, *replacements)

        if dry_run:
            continue

        # Create directories and write the file; failing if it already exists
        os.makedirs(os.path.dirname(target_fpath), exist_ok=True)
        with open(target_fpath, mode="x") as target_file:
            target_file.write(target_lines)

    print("\nFinished copying.\n")

    # Prepare for CMakeLists.txt adjustments
    cmakelists_fpath = os.path.abspath(
        os.path.join(impl_target_dir, "../CMakeLists.txt")
    )

    if not add_to_cmakelists:
        print("Not extending CMakeLists.txt automatically.")
        print(
            "Remember to register the new model in the relevant "
            f"CMakeLists.txt file at\n\t{cmakelists_fpath}\n"
            "and invoke CMake to reconfigure."
        )
        return

    print("Adding model directory to CMakeLists.txt ...")
    add_model_to_cmakelists(
        fpath=cmakelists_fpath, new_name=new_name, write=not dry_run
    )

    print("\nFinished.")



"""Test the plotting module"""

import builtins
import copy
import logging
import os
import sys

import dantro
import matplotlib as mpl
import matplotlib.pyplot as plt
import networkx as nx
import paramspace as psp
import pytest
from dantro._import_tools import remove_from_sys_modules
from dantro.data_ops import available_operations
from dantro.plot_mngr import PlotCreatorError

from utopya import Multiverse
from utopya._import_tools import added_sys_path
from utopya._import_tools import temporary_sys_modules as tmp_sys_modules
from utopya.eval.plots._graph import GraphPlot
from utopya.testtools import ModelTest
from utopya.yaml import load_yml

from .. import ADVANCED_MODEL, DUMMY_MODEL, get_cfg_fpath

# Get the test resources ......................................................
# Mute the matplotlib logger
logging.getLogger("matplotlib").setLevel(logging.WARNING)

# Basic universe plots
BASIC_UNI_PLOTS = get_cfg_fpath("plots/basic_uni.yml")

# DAG-based plots
DAG_PLOTS = get_cfg_fpath("plots/dag.yml")

# Bifurcation diagram plots, 1D and 2D
BIFURCATION_DIAGRAM_RUN = get_cfg_fpath("plots/bifurcation_diagram/run.yml")
BIFURCATION_DIAGRAM_PLOTS = get_cfg_fpath(
    "plots/bifurcation_diagram/plots.yml"
)
BIFURCATION_DIAGRAM_2D_RUN = get_cfg_fpath(
    "plots/bifurcation_diagram_2d/run.yml"
)
BIFURCATION_DIAGRAM_2D_PLOTS = get_cfg_fpath(
    "plots/bifurcation_diagram_2d/plots.yml"
)

# Graph plots
GRAPH_RUN = get_cfg_fpath("graphgroup_cfg.yml")
GRAPH_PLOTS = get_cfg_fpath("plots/graph_plot_cfg.yml")
GRAPH_PLOT_CLS = get_cfg_fpath("graphplot_class_cfg.yml")


# -----------------------------------------------------------------------------


# Fixtures --------------------------------------------------------------------
from .._fixtures import *


@pytest.fixture(autouse=True)
def register_demo_project(tmp_projects, with_test_models):
    """Use on all tests in this module"""
    pass


@pytest.fixture
def without_cached_model_plots_modules():
    remove_from_sys_modules(lambda m: m.startswith("model_plots"))

    assert "model_plots" not in sys.modules
    assert not any([ms.startswith("model_plots") for ms in sys.modules])


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


def test_dag_custom_operations(without_cached_model_plots_modules):
    """Tests if custom dantro data operations can be registered via the
    extensions made to PlotManager.
    """
    op_name = "my_custom_data_operation"

    # Make sure the operation is not yet registered
    OPERATIONS = dantro.data_ops._OPERATIONS
    if op_name in OPERATIONS:
        del OPERATIONS["my_custom_data_operation"]

    assert "my_custom_data_operation" not in available_operations()

    # Now, set up the model and its PlotManager
    model = ModelTest(ADVANCED_MODEL)
    mv, dm = model.create_run_load()
    mv.pm.raise_exc = True
    plot_cfgs = load_yml(DAG_PLOTS)

    # Should still not be available
    assert "my_custom_data_operation" not in available_operations()

    # Now, do some unrelated plot operation, i.e. not requiring model-specific
    # plot functions, but only utopya plot functions.
    # This should still lead to the _invoke_creator method being called, which
    # takes care of pre-loading model-specific module definitions, which in
    # turn call register_operation during their import
    with tmp_sys_modules():
        mv.pm.plot("test", out_dir=mv.dirs["eval"], **plot_cfgs["uni_ts"])

        assert "my_custom_data_operation" in available_operations()


def test_preloading(tmpdir, without_cached_model_plots_modules):
    """Tests the preloading feature of the utopya.PlotManager"""
    model_plot_modstr = f"model_plots.{ADVANCED_MODEL}"
    model_plot_name = "custom_plot"

    # Now create the model test object, which should load it
    model = ModelTest(ADVANCED_MODEL)
    mv, dm = model.create_run_load()
    mv.pm.raise_exc = True
    assert model_plot_modstr not in sys.modules

    mv.pm.plot_from_cfg(plot_only=(model_plot_name,))
    assert model_plot_modstr in sys.modules

    # How about a case with a bad import location; adjust a copy of the info
    # bundle to be corrupt in such a way ... Also, temporarily clear the
    # sys.path, as it _might_ include a path that allows import.
    mv.renew_plot_manager(out_dir=f"session_2/", raise_exc=False)

    del sys.modules[model_plot_modstr]
    del sys.modules["model_plots"]
    assert "model_plots" not in sys.modules
    assert model_plot_modstr not in sys.modules

    # Corrupt the bundle
    mib = mv.pm._model_info_bundle
    mpd = os.path.dirname(os.path.dirname(mib.paths["py_plots_dir"]))

    # Prepare the temporary sys.path corruption, using a valid and existing
    # path but without any content
    bad_sys_path = str(tmpdir)
    mib.paths["py_plots_dir"] = bad_sys_path

    # Need to disassociate the project to not have the PlotManager load the
    # project-defined plot directory
    mib._project_name = None

    # Assign it back
    mv.pm._model_info_bundle = mib

    # Without exception raising, this should just go ahead, even though the
    # import will fail ...
    assert not mv.pm.raise_exc
    with added_sys_path(bad_sys_path), tmp_sys_modules():
        # Remove the model plots directories from the path
        sys.path = [p for p in sys.path if not p.startswith(mpd)]

        mv.pm.plot_from_cfg(plot_only=(model_plot_name,))

        # The module will still be loaded though, because the associated
        # project will load it.
        assert model_plot_modstr in sys.modules

    # With exception raising, there should be an appropriate error message
    mv.renew_plot_manager(out_dir=f"session_3/", raise_exc=True)
    mv.pm._model_info_bundle = mib

    with added_sys_path(bad_sys_path), tmp_sys_modules():
        assert mv.pm._model_info_bundle.paths["py_plots_dir"] == bad_sys_path
        assert "model_plots" not in sys.modules
        assert model_plot_modstr not in sys.modules
        assert mv.pm.raise_exc

        remove_from_sys_modules(lambda m: m.startswith("model_plots"))
        assert not any([ms.startswith("model_plots") for ms in sys.modules])

        # This will pass, because import happens via the project's py_plots_dir
        mv.pm.plot_from_cfg(plot_only=(model_plot_name,))

        # But without the project (and also: without framework) it will fail
        mv.pm._model_info_bundle._d["project_name"] = None
        assert not mv.pm._model_info_bundle.project
        remove_from_sys_modules(lambda m: m.startswith("model_plots"))

        with pytest.raises(ImportError, match="Could not import module"):
            mv.pm.plot_from_cfg(plot_only=(model_plot_name,))


def test_plot_func_resolver_extensions():
    """Test the changes and extensions to the plot function resolver.

    ... indirectly, using some other plot creator.
    """
    model = ModelTest(ADVANCED_MODEL)
    mv, dm = model.create_run_load()
    mv.pm.raise_exc = True
    print(dm.tree)

    # Can do the default plots
    mv.pm.plot_from_cfg()

    # Can specify custom plots within utopya ... but will receive an error
    # because the data dimensionality does not match
    with pytest.raises(PlotCreatorError, match="Apply dimensionality red"):
        mv.pm.plot(
            "test",
            out_dir=mv.dirs["eval"],
            creator="universe",
            universes="all",
            module=".basic_uni",
            plot_func="lineplot",
            model_name=ADVANCED_MODEL,
            path_to_data="state",
        )

    # Cannot do a custom plot with a bad relative module import
    with pytest.raises(ModuleNotFoundError, match="Could not import"):
        mv.pm.plot(
            "test2",
            out_dir=mv.dirs["eval"],
            creator="universe",
            universes="all",
            module=".some_invalid_module",
            plot_func="lineplot",
        )

    # How about absolute imports? Nope.
    with pytest.raises(ModuleNotFoundError, match="No module named 'some'"):
        mv.pm.plot(
            "test3",
            out_dir=mv.dirs["eval"],
            creator="universe",
            universes="all",
            module="some.abs.invalid.module",
            plot_func="lineplot",
        )

    # ... and when an import fails for a custom model plot module?
    with pytest.raises(
        ModuleNotFoundError,
        match="_unrelated_ ModuleNotFoundError occurred somew",
    ):
        mv.pm.plot(
            "test4",
            out_dir=mv.dirs["eval"],
            creator="universe",
            universes="all",
            module="model_plots.some_invalid_model_name",
            plot_func="lineplot",
        )


# -----------------------------------------------------------------------------
# -- Demo model default plots -------------------------------------------------
# -----------------------------------------------------------------------------


def test_dummy_model_plotting():
    """Test plotting of the dummy model works"""
    mv, _ = ModelTest(DUMMY_MODEL).create_run_load()
    mv.pm.plot_from_cfg()


def test_advanced_model_plotting():
    """Test plotting of the dummy model works"""
    mv, _ = ModelTest(ADVANCED_MODEL).create_run_load()
    mv.pm.plot_from_cfg()


# -----------------------------------------------------------------------------
# -- CA Plots -----------------------------------------------------------------
# -----------------------------------------------------------------------------


def test_caplot():
    """Tests the plot_funcs.ca module"""
    mv, _ = ModelTest(ADVANCED_MODEL).create_run_load()

    # Run the CA plots (initial frame + animation)
    mv.pm.plot_from_cfg(plot_only=["ca/animated"])
    mv.pm.plot_from_cfg(plot_only=["ca/snapshot"])


@pytest.mark.skip("No hexagonal grid model available")
def test_caplot_hexagonal():
    """Tests the plot_funcs.ca module with hexagonal lattice"""
    update_meta_cfg = {
        "parameter_space": {
            "CopyMeGrid": {
                "cell_manager": {
                    "grid": {"structure": "hexagonal"},
                    "neighborhood": {"mode": "hexagonal"},
                }
            }
        }
    }
    mv, _ = ModelTest("CopyMeGrid").create_run_load(**update_meta_cfg)

    # Run the CA plots (initial frame + animation)
    mv.pm.plot_from_cfg(plot_only=["initial_state_and_trait"])
    mv.pm.plot_from_cfg(plot_only=["state_and_trait_anim"])

    # Same again with SimpleEG . . . . . . . . . . . . . . . . . . . . . . . .
    mv, _ = ModelTest("SimpleEG").create_run_load(**update_meta_cfg)

    # Plot the default configuration, which already includes some CA plotting
    mv.pm.plot_from_cfg()

    # To explicitly plot with the frames writer, select the disabled config
    mv.pm.plot_from_cfg(plot_only=["strategy_and_payoff_frames"])


# -----------------------------------------------------------------------------
# -- DAG Plots ----------------------------------------------------------------
# -----------------------------------------------------------------------------
# TODO Decide which ones need to be tested here at all


def test_dag_plotting():
    """Makes sure that DAG plotting works as expected"""
    # Now, set up the model
    model = ModelTest(ADVANCED_MODEL)
    mv, dm = model.create_run_load()
    mv.pm.raise_exc = True
    print(dm.tree)

    # Load some configuration arguments
    shared_kwargs = dict(out_dir=mv.dirs["eval"])
    plot_cfgs = load_yml(DAG_PLOTS)

    # Can do a simple DAG-based universe and multiverse plot
    for cfg_name, plot_cfg in plot_cfgs.items():
        if cfg_name.startswith("."):
            continue

        # The actual plotting
        print(f"Plotting '{cfg_name}' ...")
        mv.pm.plot(cfg_name, **shared_kwargs, **plot_cfg)
        print(f"Successfully plotted '{cfg_name}'!\n\n")


@pytest.mark.skip("No model with DAG based plots available")  # TODO
def test_generic_dag_plots(tmpdir):
    """Tests the plot_funcs.dag.generic module"""
    mv, _ = ModelTest("SEIRD").create_run_load()
    mv.pm.plot_from_cfg(
        plot_only=[
            "age_distribution/final",
            "age_distribution/time_series",
            "age_distribution/deceased",
        ]
    )


@pytest.mark.skip("Plot function no longer included")  # TODO
def test_time_series_plots():
    """Tests the plot_funcs.time_series module"""
    mv, _ = ModelTest("SandPile").create_run_load()

    # Plot specific plots from the default plot configuration, which are using
    # the time_series plots
    mv.pm.plot_from_cfg(plot_only=["area_fraction"])

    # Again, with PredatorPrey
    mv, _ = ModelTest("PredatorPrey").create_run_load()

    # Plot specific plots from the default plot configuration, which are using
    # the time_series plots
    mv.pm.plot_from_cfg(plot_only=["species_densities", "phase_space"])


@pytest.mark.skip("Plot function no longer included")  # TODO
def test_distribution_plots():
    """Tests the plot_funcs.distribution module"""
    mv, _ = ModelTest("SandPile").create_run_load()

    # Plot specific plots from the default plot configuration, which are using
    # the distribution plots
    mv.pm.plot_from_cfg(
        plot_only=["compl_cum_prob_dist", "cluster_size_distribution"]
    )


# -----------------------------------------------------------------------------
# -- Graph plots --------------------------------------------------------------
# -----------------------------------------------------------------------------


def test_GraphPlot_class():
    """Tests the plot_funcs._graph module.

    Mainly tests basic functionality, class attribute management and
    configuration parsing. Plotting of more complex graph data (using features
    like property-mapping) is tested in test_plotting.test_graph_plots.
    """
    graph = nx.complete_graph(3, create_using=nx.Graph)
    digraph = nx.complete_graph(3, create_using=nx.DiGraph)

    # Initialize GraphPlot with defaults
    gp = GraphPlot(graph)

    # Check that all nodes and edges are selected for drawing
    assert gp._nodes_to_draw == [0, 1, 2]
    assert len(gp._edges_to_draw) == 3
    assert gp._nodes_to_shrink == []

    # Check properties
    assert isinstance(gp.g, nx.Graph)
    assert gp.g is not gp._g

    # Draw nodes, edges, and default labels
    gp.draw(node_labels=dict(enabled=True), edge_labels=dict(enabled=True))
    # Draw again, re-creating the colormanagers
    gp.draw(nodes=dict(vmin=0.0), edges=dict(edge_vmin=0.0))

    assert isinstance(gp._mpl_nodes, mpl.collections.PathCollection)
    assert isinstance(gp._mpl_edges, mpl.collections.LineCollection)
    assert isinstance(gp._mpl_node_labels, dict)
    assert isinstance(gp._mpl_edge_labels, dict)
    assert gp._mpl_node_cbar is None
    assert gp._mpl_edge_cbar is None

    # Check that the label drawing kwargs are not set permanently
    assert gp._node_label_kwargs == {}
    assert gp._edge_label_kwargs == {}

    # Add colorbars
    gp.add_colorbars()
    # Doing this twice should be ok and should remove the first colorbars
    gp.add_colorbars()

    assert isinstance(gp._mpl_node_cbar, mpl.colorbar.Colorbar)
    assert isinstance(gp._mpl_edge_cbar, mpl.colorbar.Colorbar)

    # Clear plot
    gp.clear_plot()
    # Doing this twice should be ok
    gp.clear_plot()

    # Done with testing the drawing on this figure.
    plt.close(gp.fig)

    # Test the subgraph selection
    # Select two nodes to draw, remove 1 node and 2 edges from the graph
    gp = GraphPlot(graph, select=dict(nodelist=[0, 1]))
    assert gp._g.number_of_nodes() == 2
    assert gp._g.number_of_edges() == 1
    assert gp._nodes_to_draw == [0, 1]
    assert len(gp._edges_to_draw) == 1
    assert gp._nodes_to_shrink == []

    # Select two nodes to draw, but don't remove anything from the graph
    gp = GraphPlot(graph, select=dict(nodelist=[0, 1], drop=False))
    assert gp._g.number_of_nodes() == 3
    assert gp._g.number_of_edges() == 3
    assert gp._nodes_to_draw == [0, 1]
    assert len(gp._edges_to_draw) == 1
    assert gp._nodes_to_shrink == []

    # With open_edges, the non-selected node is still drawn but shrinked
    gp = GraphPlot(graph, select=dict(nodelist=[0, 1], open_edges=True))
    assert gp._g.number_of_nodes() == 3
    assert gp._g.number_of_edges() == 3
    assert gp._nodes_to_draw == [0, 1, 2]
    assert len(gp._edges_to_draw) == 3
    assert gp._nodes_to_shrink == [2]

    # Now go through configurations and test initialization and drawing
    configs = load_yml(GRAPH_PLOT_CLS)

    for name, cfg in configs.items():
        fig = plt.figure()

        # Try using a graphviz node layout, which requires pydot
        if name == "graphviz":
            try:
                import pydot
            except ImportError:
                with pytest.raises(
                    ImportError, match="No module named 'pydot'"
                ):
                    gp = GraphPlot(digraph, fig=fig, **cfg)

                continue

        # Configurations for which an Error is raised
        if "_raises" in cfg:
            exc_type = getattr(builtins, cfg.pop("_raises"))
            match = cfg.pop("_match")

            with pytest.raises(exc_type, match=match):
                gp = GraphPlot(digraph, fig=fig, **cfg)
                gp.draw()
                gp.clear_plot()

        # Configurations that lead to a warning
        elif "_warns" in cfg:
            warn_type = getattr(builtins, cfg.pop("_warns"))
            match = cfg.pop("_match")

            with pytest.warns(warn_type, match=match):
                gp = GraphPlot(digraph, fig=fig, **cfg)
                gp.draw()
                gp.clear_plot()

        # These should work fine
        else:
            gp = GraphPlot(digraph, fig=fig, **cfg)
            gp.draw()
            gp.clear_plot()

        plt.close(fig)


@pytest.mark.skip("No graph model available")  # TODO
def test_graph_plots(tmpdir):
    """Tests the plot_funcs.dag.graph module"""
    # Create and run simulation
    raise_exc = {"plot_manager": {"raise_exc": True}}
    mv = Multiverse(
        model_name="CopyMeGraph",
        run_cfg_path=GRAPH_RUN,
        paths=dict(out_dir=str(tmpdir)),
        **raise_exc,
    )

    mv.run_single()

    # Load
    mv.dm.load_from_cfg(print_tree=False)

    # Single graph plots
    mv.pm.plot_from_cfg(
        plots_cfg=GRAPH_PLOTS,
        plot_only=(
            "Graph",
            "DiGraph",
            "MultiGraph",
            "MultiDiGraph",
            "ExternalProperties",
            "Example_graph_plot",
            "custom_node_positioning_model",
            "explicit_node_positions",
            "custom_graph_creation",
            "custom_graph_arr_creation",
        ),
    )

    # Animation plots
    mv.pm.plot_from_cfg(
        plots_cfg=GRAPH_PLOTS,
        plot_only=[
            "graph_anim1",
            "graph_anim2",
            "graph_anim3",
            "graph_anim_external",
            "graph_anim4",
            "graph_anim_custom_graph_creation",
        ],
    )

    # Test failing cases – if possible these test are done in the (faster)
    # GraphPlot-class test.
    # Providing invalid dag tag for external property
    with pytest.raises(
        PlotCreatorError,
        match=(
            "No tag 'some_state_transformed' found in the data selected by "
            "the DAG!"
        ),
    ):
        mv.pm.plot_from_cfg(
            plots_cfg=GRAPH_PLOTS, plot_only=["invalid_ext_prop"]
        )

    # Ambiguous time specifications for animation
    with pytest.raises(
        PlotCreatorError, match="ambiguous animation time specifications"
    ):
        mv.pm.plot_from_cfg(
            plots_cfg=GRAPH_PLOTS, plot_only=["anim_amgiguous_time_spec"]
        )

    # Trying to animate from single nx.Graph
    with pytest.raises(
        PlotCreatorError, match="due to invalid type of the 'graph'"
    ):
        mv.pm.plot_from_cfg(
            plots_cfg=GRAPH_PLOTS, plot_only=["anim_not_dataarray"]
        )


# -----------------------------------------------------------------------------
# -- Bifurcation diagram plots ------------------------------------------------
# -----------------------------------------------------------------------------
# TODO Need alternative way of testing this


@pytest.mark.skip("Need alternative way of testing this")
def test_bifurcation_diagram(tmpdir):
    """Test plotting of the bifurcation diagram"""
    # Create and run simulation
    raise_exc = {"plot_manager": {"raise_exc": True}}
    mv = Multiverse(
        model_name="SavannaHomogeneous",
        run_cfg_path=BIFURCATION_DIAGRAM_RUN,
        paths=dict(out_dir=str(tmpdir)),
        **raise_exc,
    )
    mv.run_sweep()

    # Load
    mv.dm.load_from_cfg(print_tree=False)

    # Plot the bifurcation using the last datapoint
    mv.pm.plot_from_cfg(
        plots_cfg=BIFURCATION_DIAGRAM_PLOTS, plot_only=["bifurcation_endpoint"]
    )
    # Plot the bifurcation using the fixpoint
    mv.pm.plot_from_cfg(
        plots_cfg=BIFURCATION_DIAGRAM_PLOTS, plot_only=["bifurcation_fixpoint"]
    )
    mv.pm.plot_from_cfg(
        plots_cfg=BIFURCATION_DIAGRAM_PLOTS,
        plot_only=["bifurcation_fixpoint_to_plot"],
    )
    # Plot the bifurcation using scatter
    mv.pm.plot_from_cfg(
        plots_cfg=BIFURCATION_DIAGRAM_PLOTS, plot_only=["bifurcation_scatter"]
    )
    # Plot the bifurcation using oscillation
    mv.pm.plot_from_cfg(
        plots_cfg=BIFURCATION_DIAGRAM_PLOTS,
        plot_only=["bifurcation_oscillation"],
    )

    # Redo simulation, but using several initial conditions
    mv = Multiverse(
        model_name="SavannaHomogeneous",
        run_cfg_path=BIFURCATION_DIAGRAM_RUN,
        paths=dict(out_dir=str(tmpdir)),
        **raise_exc,
        parameter_space=dict(seed=psp.ParamDim(default=0, range=[4])),
    )
    mv.run_sweep()
    mv.dm.load_from_cfg(print_tree=False)

    # Plot the bifurcation using multistability
    mv.pm.plot_from_cfg(
        plots_cfg=BIFURCATION_DIAGRAM_PLOTS, plot_only=["bifurcation_fixpoint"]
    )


@pytest.mark.skip("Need alternative way of testing this")
def test_bifurcation_diagram_2d(tmpdir):
    """Test plotting of the bifurcation diagram"""
    # Create and run simulation
    raise_exc = {"plot_manager": {"raise_exc": True}}
    mv = Multiverse(
        model_name="SavannaHomogeneous",
        run_cfg_path=BIFURCATION_DIAGRAM_2D_RUN,
        paths=dict(out_dir=str(tmpdir)),
        **raise_exc,
    )
    mv.run_sweep()

    # Load
    mv.dm.load_from_cfg(print_tree=False)

    # Plot the bifurcation using the last datapoint
    mv.pm.plot_from_cfg(
        plots_cfg=BIFURCATION_DIAGRAM_2D_PLOTS,
        plot_only=["bifurcation_diagram_2d"],
    )
    # Plot the bifurcation using the fixpoint
    mv.pm.plot_from_cfg(
        plots_cfg=BIFURCATION_DIAGRAM_2D_PLOTS,
        plot_only=["bifurcation_diagram_2d_fixpoint_to_plot"],
    )



"""Test the plot_funcs._mpl_helpers module"""
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pytest
from pkg_resources import resource_filename

from utopya.eval.plots._mpl import ColorManager
from utopya.tools import load_yml

COLOR_MANAGER_CFG = resource_filename("tests", "cfg/color_manager_cfg.yml")

# Fixtures --------------------------------------------------------------------


# Tests -----------------------------------------------------------------------


def test_ColorManager():
    """Tests the ColorManager class."""
    test_configurations = load_yml(COLOR_MANAGER_CFG)

    # Test initializing the default ColorManager
    colormanager = ColorManager()
    assert colormanager.cmap.name == "viridis"
    assert isinstance(colormanager.norm, mpl.colors.Normalize)
    assert colormanager.labels is None

    for name, cfg in test_configurations.items():
        cbar_kwargs = cfg.pop("cbar_kwargs", {})

        # Test the failing cases
        if name == "invalid_norm":
            with pytest.raises(
                ValueError, match="Received invalid norm specifier:"
            ):
                ColorManager(**cfg)
            continue

        if name == "invalid_cmap":
            with pytest.raises(
                ValueError, match="Received invalid colormap name:"
            ):
                ColorManager(**cfg)
            continue

        if name == "disconnected_intervals":
            with pytest.raises(
                ValueError, match="Received disconnected intervals:"
            ):
                ColorManager(**cfg)
            continue

        if name == "decreasing_boundaries":
            with pytest.raises(
                ValueError, match="Received decreasing boundaries:"
            ):
                ColorManager(**cfg)
            continue

        if name == "single_bin_center":
            with pytest.raises(
                ValueError, match="At least 2 bin centers must be given"
            ):
                ColorManager(**cfg)
            continue

        # Initialize the ColorManager and retreive the created colormap, norm,
        # and colorbar labels.
        colormanager = ColorManager(**cfg)

        cmap = colormanager.cmap
        norm = colormanager.norm
        labels = colormanager.labels

        assert isinstance(cmap, mpl.colors.Colormap)
        assert isinstance(norm, mpl.colors.Normalize)
        assert labels is None or isinstance(labels, dict)

        # Test the `ColorManager.create_cbar` method
        fig, ax = plt.subplots()
        scatter = ax.scatter(
            np.arange(10), np.arange(10), c=np.arange(10), cmap=cmap, norm=norm
        )

        cb = colormanager.create_cbar(scatter, **cbar_kwargs)

        assert isinstance(cb, mpl.colorbar.Colorbar)
        assert cb.norm == norm
        assert cb.cmap == cmap
        if "labels" in cfg and isinstance(cfg["labels"], dict):
            assert (cb.get_ticks() == list(cfg["labels"].keys())).all()
        plt.close()

        # Test the `ColorManager.map_to_color` method
        colors = colormanager.map_to_color(42.0)
        assert mpl.colors.is_color_like(colors)
        colors = colormanager.map_to_color(np.linspace(2.1, 5.3, 10))
        assert all([mpl.colors.is_color_like(c) for c in colors])

        # NOTE Some explicit color checks. If it fails check the respective
        #      configurations.
        if name == "from_intervals":
            assert cmap(1) == mpl.colors.to_rgba("w")

        if name == "shortcut_categorical":
            assert cmap(0) == mpl.colors.to_rgba("g")

    # Test ColorManager when passing norm and cmap (as mpl object) directly
    colormanager = ColorManager(
        cmap=mpl.colors.ListedColormap(["r", "b"]),
        norm=mpl.colors.BoundaryNorm([-2, 1, 2], ncolors=2),
    )
    assert colormanager.cmap(0) == mpl.colors.to_rgba("r")



"""DEPRECATED module that provides backwards-compatibility for the old utopya
module structure.

.. deprecated:: 1.0.0
"""

from .eval import MultiverseGroup, PlotManager, UniverseGroup
from .eval.plotcreators import MultiversePlotCreator, UniversePlotCreator
from .eval.plothelper import PlotHelper
from .eval.plots import is_plot_func
from .eval.transform import register_operation
